import * as path from 'path'
import * as fs from 'fs'
import * as mkdirp from 'mkdirp'
import YAML from 'yaml'
import { DataObject, DataObjectFn, Storage } from './storage.interface'

export class FolderStorage implements Storage {
  private static Folder = '.evalassist'
  private static initFiles: DataObjectFn[] = []

  static registerInitFiles(fn: DataObjectFn) {
    FolderStorage.initFiles.push(fn)
  }

  private get folder(): string {
    if (!this.projectPath.endsWith(FolderStorage.Folder) || !this.projectPath.endsWith(FolderStorage.Folder + '/')) {
      return path.join(this.projectPath, FolderStorage.Folder)
    }
    return this.projectPath
  }

  constructor(private projectPath: string) {
    // Create storage folder
    if (!fs.existsSync(this.folder)) {
      mkdirp.sync(this.folder)
    }
  }

  list(key: string): string[] {
    const path: string = this.getKeyPath(key)
    return fs.readdirSync(path)
  }

  clear(key: string): boolean {
    const path = this.getKeyPath(key)
    if (!fs.existsSync(path)) {
      return false
    } else {
      fs.unlinkSync(path)
      return true
    }
  }

  private getKeyPath(key: string) {
    return this.getKeyPathRaw(key) + '.yml'
  }

  private getKeyPathRaw(key: string) {
    return path.join(this.folder, key)
  }

  private getKeyPathFolder(keyPath: string) {
    return keyPath.substr(0, keyPath.lastIndexOf('/'))
  }

  private serialize(data: any): string {
    return YAML.stringify(data)
  }

  private unserialize<T>(data: string): T {
    return YAML.parse(data) as T
  }

  get<T>(key: string): T | null {
    const path = this.getKeyPath(key)

    // Does not exist
    if (!fs.existsSync(path)) return null

    return this.unserialize(fs.readFileSync(path).toString())
  }

  set(...dataObjects: DataObject[]) {
    for (const dataObject of dataObjects) {
      const path = this.getKeyPath(dataObject.key)
      const folder = this.getKeyPathFolder(path)

      // Create parent folders
      if (!fs.existsSync(folder)) mkdirp.sync(folder)

      // Write
      fs.writeFileSync(path, this.serialize(dataObject.data))
    }
  }

  getRaw(key: string): string | null {
    const path = this.getKeyPathRaw(key)

    // Does not exist
    if (!fs.existsSync(path)) return null

    return fs.readFileSync(path).toString()
  }

  setRaw(key: string, data: string) {
    const path = this.getKeyPathRaw(key)
    const folder = this.getKeyPathFolder(path)

    // Create parent folders
    if (!fs.existsSync(folder)) mkdirp.sync(folder)

    fs.writeFileSync(path, data)
  }

  setFolder(key: string) {
    const path = this.getKeyPathRaw(key)
    const folder = this.getKeyPathFolder(path)

    // Create parent folders
    if (!fs.existsSync(folder)) mkdirp.sync(folder)
  }
}
