import { promisify } from 'util'
import { StorageProvider, Storage } from '../../core/storage'
import { EvaluationStorageReader } from '../evaluation/storage/evaluation.storage.reader'
import { EvaluationStorageWriter } from '../evaluation/storage/evaluation.storage.writer'
import { resolve } from 'path'
import { exec } from 'child_process'
import { ProjectService } from '../project'
import { DateProvider } from '../../core/date'

export class DockerService {
  constructor(
    private storageProvider: StorageProvider,
    private projectService: ProjectService,
    private dateProvider: DateProvider,
  ) {}

  async run(
    projectPath: string,
    evaluationName: string,
    opts: { [key: string]: any },
    overwriteAnalysis: boolean,
    overwriteScore: boolean,
    dockerOpts?: any,
  ): Promise<string> {
    if (opts.docker) {
      const writer = this.getEvaluationDataWriter(projectPath)

      let cmd = 'docker run -v ' + projectPath + ':/Project '

      if (dockerOpts['volumes']) {
        for (let i = 0; i < dockerOpts['volumes'].length; i++) {
          const splitted = dockerOpts['volumes'][i].split(':/', 2)
          const basePath = resolve(splitted[0]) + ':/' + splitted[1]
          cmd = cmd + '-v ' + basePath + ' '
        }
      }
      if (dockerOpts['envs']) {
        for (let i = 0; i < dockerOpts['envs'].length; i++) {
          cmd = cmd + '-e ' + dockerOpts['envs'][i] + ' '
        }
      }

      const reader = this.getEvaluationDataReader(projectPath)
      let scoreOnly = opts.scoreOnly
      let analyzeOnly = opts.analyzeOnly
      if (!scoreOnly) {
        if (!overwriteAnalysis && reader.readAnalysisResult(evaluationName) !== null) {
          scoreOnly = true
          console.log('Analysis already exists, skipping.')
        }
      }
      if (!analyzeOnly) {
        if (!overwriteScore && reader.readScore(evaluationName) !== null) {
          analyzeOnly = true
          console.log('Score already exists, skipping.')
        }
      }
      cmd =
        cmd +
        '-e evaluationName=' +
        evaluationName +
        (scoreOnly == true ? ' -e scoreOnly=true' : '') +
        (analyzeOnly == true ? ' -e analyzeOnly=true' : '') +
        ' ' +
        opts.docker

      console.log("Currently running '" + cmd + "'")

      cmd = cmd + ' 2>&1'

      // If Analysis and Score files already exists, skipping all, scoreOnly and analyzeOnly options are already processed in CLI
      if (!(scoreOnly == true && analyzeOnly == true)) {
        writer.createTmp()
        const res = await dockerCall(cmd)
        console.log(res)
        return res
      }
      return 'nothing passed !'
    }
    console.log('nothing passed !')
    return 'no image !'
  }

  private getEvaluationDataReader(projectPath: string) {
    const storage: Storage = this.storageProvider.forPath(projectPath)
    return new EvaluationStorageReader(storage)
  }

  private getEvaluationDataWriter(projectPath: string) {
    const storage = this.storageProvider.forPath(projectPath)
    return new EvaluationStorageWriter(storage)
  }

  /**
   * Read the analysis of the project, and stores the results on storage.
   * @param projectPath The path of the project.
   * @param evaluationName The name of the performed analysis, to load/store the results.
   * @param overwrite If true, an existing analysis results will be overwritten.
   * @returns True is a new analysis was done, false if nothing was done (eg. the project was already scored).
   */
  async analyzeWithStorage(projectPath: string, evaluationName: string, overwrite = false): Promise<boolean> {
    const writer = this.getEvaluationDataWriter(projectPath)
    const reader = this.getEvaluationDataReader(projectPath)
    if (!overwrite && reader.readAnalysisResult(evaluationName) !== null) {
      return false
    }
    const result = reader.readDockerAnalysisResult()
    result.date = this.dateProvider.getCurrentDate()
    result.commit = this.projectService.getProjectDataReader(projectPath).lastCommitData().hash
    writer.writeAnalysisResult(evaluationName, result)
    return true
  }

  /**
   * Read the analysis of the project, and stores the results on storage.
   * @param projectPath The path of the project to score.
   * @param evaluationName The name of the performed score, to load/store the results.
   * @param overwrite If true, an existing scoring will be overwritten.
   * @returns True is a new scoring was done, false if nothing was done (eg. the project was already scored).
   */
  async scoreWithStorage(projectPath: string, evaluationName: string, overwrite = false): Promise<boolean> {
    const writer = this.getEvaluationDataWriter(projectPath)
    const reader = this.getEvaluationDataReader(projectPath)
    if (!overwrite && reader.readScore(evaluationName) !== null) {
      return false
    }
    const result = reader.readDockerScore()
    result.date = this.dateProvider.getCurrentDate()
    result.commit = this.projectService.getProjectDataReader(projectPath).lastCommitData().hash
    writer.writeScore(evaluationName, result)
    return true
  }

  async deleteDocker(projectPath: string) {
    const writer = this.getEvaluationDataWriter(projectPath)
    writer.deleteDocker()
  }
}

async function dockerCall(command: string): Promise<string> {
  const execute = promisify(exec)
  const res = await execute(command, {
    cwd: '.',
  })
  return res.stdout
}
