import { Evaluator, CSVProjectLine, CSVMemberLine } from '.'
import { CSVMaker } from '../../core/csv'
import { StorageProvider, Storage } from '../../core/storage'
import { Project, ProjectService } from '../project'
import { Member, MemberService } from '../member'
import { EvaluationStorageReader } from './storage/evaluation.storage.reader'
import { EvaluationStorageWriter } from './storage/evaluation.storage.writer'
import { FileWriter } from '../../core/file'
import { Problem } from '../../core/exceptions'
import { DateProvider } from '../../core/date'
import * as fs from 'fs'
import * as YAML from 'yaml'
import { exec } from 'child_process'
import { promisify } from 'util'

export class EvaluationService {
  constructor(
    private storageProvider: StorageProvider,
    private projectService: ProjectService,
    private memberService: MemberService,
    private csvWriter: CSVMaker,
    private fileWriter: FileWriter,
    private dateProvider: DateProvider,
  ) {}

  getEvaluationDataReader(projectPath: string) {
    const storage: Storage = this.storageProvider.forPath(projectPath)
    return new EvaluationStorageReader(storage)
  }

  private getEvaluationDataWriter(projectPath: string) {
    const storage = this.storageProvider.forPath(projectPath)
    return new EvaluationStorageWriter(storage)
  }

  private async getNpmGlobalRoot(): Promise<string> {
    const execAsync = promisify(exec)
    const { stdout: globalPath } = await execAsync('npm root -g')
    return globalPath.trim()
  }

  async loadEvaluatorWithGlobalRetry(evaluatorPath: string, evaluatorConfiguration?: any): Promise<Evaluator> {
    try {
      return await this.loadEvaluator(evaluatorPath, evaluatorConfiguration)
    } catch (e) {
      const globalPath = `${await this.getNpmGlobalRoot()}/${evaluatorPath}`
      console.log('Could not find the evaluator locally, trying with npm globally installed packages.')
      return await this.loadEvaluator(globalPath, evaluatorConfiguration)
    }
  }

  private async loadEvaluator(evaluatorPath: string, evaluatorConfiguration?: any): Promise<Evaluator> {
    try {
      const importedContent = await import(evaluatorPath)
      for (const key of Object.keys(importedContent)) {
        if (key.includes('Evaluator')) {
          const importedClass = importedContent[key]
          const evaluator: Evaluator = new importedClass()
          if (evaluatorConfiguration && evaluator.initialize) {
            evaluator.initialize(evaluatorConfiguration!)
          }
          console.log('Successfully loaded evaluator from ' + evaluatorPath)
          return evaluator
        }
      }
      throw new Problem(`The provided module does not contain a member with the "Evaluator" suffix.`)
    } catch (e) {
      throw new Problem(`Could not load evaluator from '${evaluatorPath}' because of the following error: "${e}"`)
    }
  }

  loadEvaluatorConfiguration(evaluatorConfigurationPath: string): any {
    const configurationYAML: string = fs.readFileSync(evaluatorConfigurationPath, { encoding: 'utf-8' })
    const configuration = YAML.parse(configurationYAML)
    return configuration
  }

  isAnalysisDataOutdated(projectPath: string, evaluationName: string): boolean {
    const reader = this.getEvaluationDataReader(projectPath)
    const currentCommit = this.projectService.getProjectDataReader(projectPath).lastCommitData().hash
    const previousResults = reader.readAnalysisResult(evaluationName)
    if (previousResults) {
      const previousCommit: string = previousResults.commit
      return previousCommit != currentCommit
    } else {
      return true
    }
  }

  /**
   * Analyzes a project, and stores the results on storage.
   * @param projectPath The path of the project to analyze.
   * @param evaluatorPath The path of the evaluator to use.
   * @param evaluationName The name of the performed analysis, to load/store the results.
   * @param overwrite If true, an existing analysis results will be overwritten.
   * @returns True is a new analysis was done, false if nothing was done (eg. the project was already scored).
   */
  async analyzeWithStorage(
    projectPath: string,
    evaluator: Evaluator,
    evaluationName: string,
    overwrite = false,
  ): Promise<boolean> {
    const writer = this.getEvaluationDataWriter(projectPath)
    const reader = this.getEvaluationDataReader(projectPath)
    if (!overwrite && reader.readAnalysisResult(evaluationName) !== null) {
      return false
    }
    const result = await evaluator.analyze(projectPath)
    result.date = this.dateProvider.getCurrentDate()
    result.commit = this.projectService.getProjectDataReader(projectPath).lastCommitData().hash
    writer.writeAnalysisResult(evaluationName, result)
    return true
  }

  isScoringDataOutdated(projectPath: string, evaluationName: string): boolean {
    const reader = this.getEvaluationDataReader(projectPath)
    const currentCommit = this.projectService.getProjectDataReader(projectPath).lastCommitData().hash
    const previousScore = reader.readScore(evaluationName)
    if (previousScore) {
      const previousCommit: string = previousScore.commit
      return previousCommit != currentCommit
    } else {
      return true
    }
  }

  /**
   * Use the analysis results from storage to computes the score of a project, and stores this score on storage.
   * @param projectPath The path of the project to score.
   * @param evaluatorPath The path of the evaluator to use.
   * @param evaluationName The name of the performed analysis, to load/store the results.
   * @param overwrite If true, an existing scoring will be overwritten.
   * @returns True is a new scoring was done, false if nothing was done (eg. the project was already scored).
   */
  async scoreWithStorage(
    projectPath: string,
    evaluator: Evaluator,
    evaluationName: string,
    overwrite = false,
  ): Promise<boolean> {
    const writer = this.getEvaluationDataWriter(projectPath)
    const reader = this.getEvaluationDataReader(projectPath)

    if (reader.readScore(evaluationName) && !overwrite) {
      return false
    }

    const result = reader.readAnalysisResult(evaluationName)
    if (result === null) {
      throw new Problem(`There are no analysis results named '${evaluationName}' in this project.`)
    }
    const evaluationResults = await evaluator.score(result)
    evaluationResults.date = this.dateProvider.getCurrentDate()
    evaluationResults.commit = this.projectService.getProjectDataReader(projectPath).lastCommitData().hash
    writer.writeScore(evaluationName, evaluationResults)
    return true
  }

  private generateCSVData(
    projectsFolders: string[],
    evaluationName: string,
    membersToHide: string[],
    extractCSVLines: (score: any, projectData: Project, members: Member[]) => Array<CSVMemberLine | CSVProjectLine>,
  ): Array<CSVMemberLine | CSVProjectLine> {
    let csvLines: Array<CSVMemberLine | CSVProjectLine> = []
    for (const folderFullPath of projectsFolders) {
      csvLines = csvLines.concat(
        this.generateCSVDataSingle(folderFullPath, evaluationName, membersToHide, extractCSVLines),
      )
    }
    return csvLines
  }

  private generateCSVDataSingle(
    folderFullPath: string,
    evaluationName: string,
    membersToHide: string[],
    extractCSVLines: (score: any, projectData: Project, members: Member[]) => Array<CSVMemberLine | CSVProjectLine>,
  ): Array<CSVMemberLine | CSVProjectLine> {
    const evaluationReader = this.getEvaluationDataReader(folderFullPath)
    const projectReader = this.projectService.getProjectDataReader(folderFullPath)
    const memberReader = this.memberService.getProjectDataReader(folderFullPath)
    const score: any = evaluationReader.readScore(evaluationName)
    if (score === null) {
      throw new Problem(`There are no scoring results to use in the project: ${folderFullPath}.`)
    }
    const projectData: Project = projectReader.project()
    const members: Member[] = memberReader
      .members()
      .filter((m1) => !membersToHide.map((m2) => m2.toLowerCase()).includes(m1.username.toLowerCase()))
    return extractCSVLines(score, projectData, members)
  }

  extractMembersCSVData(score: any, projectData: Project, members: Member[]): Array<CSVMemberLine> {
    return members.map((member) => {
      return {
        projectID: projectData.id,
        memberID: member.username,
        memberName: member.name,
        data: score,
      }
    })
  }

  extractProjectCSVData(score: any, projectData: Project, members: Member[]): Array<CSVProjectLine> {
    return [
      {
        projectID: projectData.id,
        members: members.map((m) => m.name).join(' - '),
        membersIDs: members.map((m) => m.username).join(' - '),
        data: score,
      },
    ]
  }

  generatePerMemberCSVDataSingle(
    projectFolder: string,
    evaluationName: string,
    membersToHide: string[] = [],
  ): Array<CSVMemberLine | CSVProjectLine> {
    return this.generateCSVDataSingle(projectFolder, evaluationName, membersToHide, this.extractMembersCSVData)
  }

  generatePerProjectCSVDataSingle(
    projectFolder: string,
    evaluationName: string,
    membersToHide: string[] = [],
  ): Array<CSVMemberLine | CSVProjectLine> {
    return this.generateCSVDataSingle(projectFolder, evaluationName, membersToHide, this.extractProjectCSVData)
  }

  generatePerMemberCSVData(
    projectsFolders: string[],
    evaluationName: string,
    membersToHide: string[] = [],
  ): Array<CSVMemberLine | CSVProjectLine> {
    const result = this.generateCSVData(projectsFolders, evaluationName, membersToHide, this.extractMembersCSVData)
    return result
  }

  generatePerProjectCSVData(
    projectsFolders: string[],
    evaluationName: string,
    membersToHide: string[] = [],
  ): Array<CSVMemberLine | CSVProjectLine> {
    return this.generateCSVData(projectsFolders, evaluationName, membersToHide, this.extractProjectCSVData)
  }

  generatePerMemberCSVFile(
    projectsFolders: string[],
    evaluationName: string,
    outputFile: string,
    membersToHide: string[] = [],
  ): void {
    const data = this.generatePerMemberCSVData(projectsFolders, evaluationName, membersToHide)
    const csvContent = this.csvWriter.toCSVWithDefaultDelimiters(data)
    this.fileWriter.writeFileSync(outputFile, csvContent)
  }

  generatePerProjectCSVFile(
    projectsFolders: string[],
    evaluationName: string,
    outputFile: string,
    membersToHide: string[] = [],
  ): void {
    const data = this.generatePerProjectCSVData(projectsFolders, evaluationName, membersToHide)
    const csvContent = this.csvWriter.toCSVWithDefaultDelimiters(data)
    this.fileWriter.writeFileSync(outputFile, csvContent)
  }

  deleteEvaluationData(projectFolder: string, evaluationName: string): boolean {
    const writer = this.getEvaluationDataWriter(projectFolder)
    const deleted1 = writer.deleteAnalysisResult(evaluationName)
    const deleted2 = writer.deleteScore(evaluationName)
    return deleted1 || deleted2
  }
}
