import { EvaluationService } from '.'
import { CSVMaker, DefaultCSVMaker } from '../../core/csv'
import { DateProvider, DefaultDateProvider } from '../../core/date'
import { FileWriter } from '../../core/file'
import { HTTPClient } from '../../core/http-client'
import { DefaultStorageProvider, Storage } from '../../core/storage'
import { CommitData, Project, ProjectService } from '../project'
import { Member, MemberService } from '../member'
import { Evaluator } from './evaluation.interface'

let instance: EvaluationService
let mockSet: jest.Mock
let mockGet: jest.Mock
let mockSetRaw: jest.Mock
let mockWrite: jest.Mock
let mockGetCurrentDate: jest.Mock
const fakeData = { theKey: 'Some data' }
const fakeResults = { lettersIntheKey: fakeData.theKey.length }

// Project
const dummyProject: Project = {
  id: 1,
  name: 'Dummy project',
  namespacePath: '',
  description: '',
  webUrl: '',
  createdAt: {} as Date,
}

// Members
const member1: Member = {
  id: 2,
  name: 'Alice',
  username: 'E2',
  url: '',
}
const member2: Member = {
  id: 3,
  name: 'Bob',
  username: 'E3',
  url: '',
}
const members: Member[] = [member1, member2]

let mockEvaluator: Evaluator

describe('Evaluation Service', () => {
  beforeEach(() => {
    // Mock the storage
    mockSet = jest.fn()
    mockGet = jest.fn()
    mockSetRaw = jest.fn()
    const mockStorageProvider = {
      forPath(path: string): Storage {
        return {
          clear: jest.fn(),
          get: mockGet,
          getRaw: jest.fn(),
          set: mockSet,
          setRaw: mockSetRaw,
          setFolder: jest.fn(),
          list: jest.fn(),
        }
      },
    }

    mockWrite = jest.fn()
    const mockFileWriter: FileWriter = {
      writeFileSync: mockWrite,
    }

    mockGetCurrentDate = jest.fn()
    const mockDateProvider: DateProvider = {
      getCurrentDate: mockGetCurrentDate,
    }

    // Prepare instance
    instance = new EvaluationService(
      mockStorageProvider,
      new ProjectService({} as HTTPClient, mockStorageProvider),
      new MemberService({} as HTTPClient, mockStorageProvider),
      new DefaultCSVMaker(),
      mockFileWriter,
      mockDateProvider,
    )

    // Mock the evaluator
    mockEvaluator = {
      analyze(path: string): any {
        return fakeData
      },
      score(analysisResults: any): any {
        const data: string = analysisResults.someKey
        return fakeResults
      },
    }
  })

  it('loadEvaluator with a dummy evaluator', async () => {
    const normalInstance = new EvaluationService(
      new DefaultStorageProvider(),
      {} as ProjectService,
      {} as MemberService,
      {} as CSVMaker,
      {} as FileWriter,
      {} as DateProvider,
    )
    const dummyEvaluator: Evaluator = await normalInstance.loadEvaluatorWithGlobalRetry('./test-data/dummy-evaluator')
    expect(dummyEvaluator).toBeDefined()
    expect(dummyEvaluator.analyze).toBeDefined()
    expect(dummyEvaluator.score).toBeDefined()
  })

  it('analyze with a mock evaluator', async () => {
    mockGet.mockReturnValueOnce(null)
    mockGet.mockReturnValueOnce({ hash: 'someHash' })
    mockGetCurrentDate.mockReturnValueOnce(new Date(1337))

    // Operation call
    await instance.analyzeWithStorage('test-data', mockEvaluator, 'dummyAnalysis')

    // Oracle
    expect(mockSet.mock.calls[0][0]).toMatchSnapshot()
  })

  it('score with a mock evaluator', async () => {
    mockGet.mockReturnValueOnce(null)
    mockGet.mockReturnValueOnce(fakeData)
    mockGet.mockReturnValueOnce({ hash: 'someHash' })
    mockGetCurrentDate.mockReturnValueOnce(new Date(1337))

    // Operation call
    await instance.scoreWithStorage('test-data', mockEvaluator, 'dummyAnalysis')

    // Oracle
    expect(mockSet.mock.calls[0][0]).toMatchSnapshot()
  })

  it('generatePerProjectCSVFile with mocks', async () => {
    // Mocking the retrived data
    mockGet.mockReturnValueOnce(fakeResults)
    mockGet.mockReturnValueOnce(dummyProject)
    mockGet.mockReturnValueOnce(members)

    // Operation call
    instance.generatePerProjectCSVFile(['test-data'], 'dummyAnalysis', 'outputFile.csv')

    // Oracle
    expect(mockWrite.mock.calls[0][0]).toMatchSnapshot()
    expect(mockWrite.mock.calls[0][1]).toMatchSnapshot()
  })

  it('generatePerMemberCSVFile with mocks', async () => {
    // Mocking the retrived data
    mockGet.mockReturnValueOnce(fakeResults)
    mockGet.mockReturnValueOnce(dummyProject)
    mockGet.mockReturnValueOnce(members)

    // Operation call
    instance.generatePerMemberCSVFile(['test-data'], 'dummyAnalysis', 'outputFile.csv')

    // Oracle
    expect(mockWrite.mock.calls[0][0]).toMatchSnapshot()
    expect(mockWrite.mock.calls[0][1]).toMatchSnapshot()
  })
})
