import { HTTPClient, HTTPResponse } from '../../core/http-client'
import { Feedback } from './feedback.interface'
import { StorageProvider } from '../../core/storage'
import { FeedbackStorageReader, FeedbackStorageWriter } from './storage'

export class FeedbackService {
  constructor(private client: HTTPClient, private storageProvider: StorageProvider) {}

  getProjectDataReader(projectPath: string) {
    const storage = this.storageProvider.forPath(projectPath)
    return new FeedbackStorageReader(storage)
  }

  getProjectDataWriter(projectPath: string) {
    const storage = this.storageProvider.forPath(projectPath)
    return new FeedbackStorageWriter(storage)
  }

  async submitFeedback(projectId: number, feedback: Feedback) {
    const splitMessage =
      '\n\n[**Please keep reading the next comment** − original message was too long and cut in several parts]'
    const sizeLimit: number = 5000 - splitMessage.length
    const messages: string[] = []
    let remainingText: string = feedback.description

    // While we are above the size limit, we cut
    while (remainingText.length > sizeLimit) {
      // Find a nice place where to cut
      const maxFirstPart: string = remainingText.substring(0, sizeLimit)
      let cutIndex: number = maxFirstPart.lastIndexOf('\n\n')
      if (cutIndex == -1) {
        cutIndex = maxFirstPart.lastIndexOf('\n')
      }
      if (cutIndex == -1) {
        cutIndex = sizeLimit
      }

      // Cut the text and add split message
      messages.push(remainingText.substring(0, cutIndex) + splitMessage)
      remainingText = remainingText.substring(cutIndex)
    }

    // Add the remaining text as last message
    messages.push(remainingText)

    // Use the first message to create the issue
    const answer: HTTPResponse<any> = await this.client.post(`/projects/${projectId}/issues`, {
      params: {
        title: feedback.title,
        description: messages.splice(0, 1)[0],
      },
    })

    // Retrieve created issue iid
    const issueIid = answer.data.iid

    // Use other messages to create additional comments
    for (const nextPart of messages) {
      await this.client.post(`/projects/${projectId}/issues/${issueIid}/notes`, {
        params: {
          body: nextPart,
        },
      })
    }
  }

  create(path: string, name: string, content: string) {
    this.getProjectDataWriter(path).writePending(name, content)
  }

  // Mark a feedback as submitted
  markFeedbackAsSubmitted(projectPath: string, feedbackName: string) {
    const reader = this.getProjectDataReader(projectPath)
    const content = reader.readPending(feedbackName)
    if (!content) return

    const writer = this.getProjectDataWriter(projectPath)
    writer.deletePending(feedbackName)
    writer.writeSubmitted(feedbackName, content)
  }
}
