import { HTTPClient } from '../../core/http-client'
import { promisify } from 'util'
import { exec } from 'child_process'
import { CloneProjectParams, OnGitChanges } from './git.interface'
import * as path from 'path'
import { ProjectService } from '../project'
import { Problem } from '../../core/exceptions'

export class GitService {
  constructor(
    private client: HTTPClient,
    private projectService: ProjectService,
    private onProjectCloneListeners: OnGitChanges[] = [],
  ) {}

  addListener(listener: OnGitChanges) {
    this.onProjectCloneListeners.push(listener)
  }

  // Clone a project
  async clone({ projectId, projectUrl, folderName }: CloneProjectParams) {
    // Clone
    const urlSplit = projectUrl.split('//')
    const cloneUrl = `${urlSplit[0]}//oauth2:${this.client.config.accessToken}@${urlSplit[1]}`
    await execGitCommandSync(`git clone ${cloneUrl} ${folderName}`)

    // Callbacks
    const clonePath = path.join(process.cwd(), folderName)
    for (const listener of this.onProjectCloneListeners) {
      await listener.onProjectClone({ projectId: projectId, path: clonePath })
    }
  }

  async pull(clonePath: string): Promise<boolean> {
    let didSomething = true
    const output = await execGitCommandSync('LANG=en git pull', clonePath)
    if (output.includes('Already up to date')) {
      didSomething = false
    }

    const id = this.projectService.getProjectDataReader(clonePath).project().id
    for (const listener of this.onProjectCloneListeners) {
      await listener.onProjectPull({ projectId: id, path: clonePath })
    }
    return didSomething
  }
}

async function execGitCommandSync(command: string, workingFolder = '.'): Promise<string> {
  const execute = promisify(exec)
  const commandOutputs = await execute(command, {
    cwd: workingFolder,
  })
  if (commandOutputs.stderr.includes('fatal')) {
    throw new Problem(`An error occured while running '${command}': ` + commandOutputs.stderr)
  }
  return commandOutputs.stdout
}
