import { HTTPClient } from '../../core/http-client'
import { StorageProvider } from '../../core/storage'
import {
  ProjectListParams,
  ProjectCollection,
  Project,
  Pagination,
  DeadlineCheckResult,
  CommitData,
} from './project.interface'
import { ProjectStorageReader } from './storage'

export class ProjectService {
  constructor(private client: HTTPClient, private storageProvider: StorageProvider) {}

  getProjectDataReader(projectPath: string) {
    const storage = this.storageProvider.forPath(projectPath)
    return new ProjectStorageReader(storage)
  }

  async get(projectId: number): Promise<Project> {
    const { data } = await this.client.get(`/projects/${projectId}`)
    return new Project(data)
  }

  async list(params: ProjectListParams = {}): Promise<ProjectCollection> {
    const { search, page, per_page, membership, date, userID, projectURL, projectID } = params

    let requestUrl = '/projects'
    let requestParams: any = {}

    // If single project, then simple request URL using the project ID
    if (projectID || projectURL) {
      let cleanProjectID = ''
      if (projectID) {
        cleanProjectID = projectID
      } else if (projectURL) {
        cleanProjectID = projectURL.replace(this.client.config.host!, '')
        while (cleanProjectID.startsWith('/')) {
          cleanProjectID = cleanProjectID.substring(1, cleanProjectID.length - 1)
        }
      }
      requestUrl = `${requestUrl}/${cleanProjectID.replace(/\//gi, '%2F')}`
    }

    // Else, prepare params
    else {
      if (userID) requestUrl = `users/${userID}/projects`
      requestParams = {
        params: { search, page, per_page, membership },
      }
    }

    // Api Call
    const { data, headers } = await this.client.get<any[]>(requestUrl, requestParams)

    // Unserialize Projects
    let projects = data.map ? data.map((x) => new Project(x)) : [new Project(data)]

    // Filter by date
    if (projects.length > 0 && date && date > projects[projects.length - 1].createdAt) {
      const lastValidPos = projects.findIndex((x) => x.createdAt < date)
      projects = projects.slice(0, lastValidPos)
    }

    return { pagination: new Pagination(headers), data: projects }
  }

  async getLastCommitData(projectId: number): Promise<CommitData | null> {
    const { data } = await this.client.get<any[]>(`projects/${projectId}/repository/commits`, {
      params: { per_page: 1 },
    })
    if (data.length) {
      const date = new Date(data[0]['committed_date'])
      const hash = data[0]['id']
      const author = data[0]['author_name']
      return { date: date, hash: hash, author: author }
    } else {
      return null
    }
  }

  checkDeadline(lastCommitDate: Date, deadline: Date): DeadlineCheckResult {
    // Positive: late
    // Negative: not late
    const diff = lastCommitDate.getTime() - deadline.getTime()
    const result: DeadlineCheckResult = { isLate: diff > 0 }
    if (result.isLate) {
      const diffSeconds = Math.floor(diff / 1000)
      const diffMinutes = Math.floor(diffSeconds / 60)

      const diffHours = Math.floor(diffMinutes / 60)
      const extraMinutes = diffMinutes % 60

      const diffDays = Math.floor(diffHours / 24)
      const extraHours = diffHours % 24

      result.days = diffDays
      result.hours = extraHours
      result.minutes = extraMinutes
    }

    return result
  }

  checkDeadlineWithStorage(path: string, deadline: Date): DeadlineCheckResult {
    const reader = this.getProjectDataReader(path)
    const lastCommitDate = new Date(reader.lastCommitData().date)
    return this.checkDeadline(lastCommitDate, deadline)
  }
}
