import { AxiosHTTPClient, HTTPClientConfig } from './core/http-client'
import { DefaultStorageProvider, StorageProvider } from './core/storage'
import { GitService } from './domain/git/git.service'
import { DockerService } from './domain/docker/docker.service'
import { ProjectDataProvider } from './domain/project'
import { ProjectService } from './domain/project/project.service'
import { ProjectDataCollectWriter } from './domain/projectData'
import { ProjectMembersDataProvider } from './domain/member'
import { MemberService } from './domain/member/member.service'
import { FeedbackService } from './domain/feedback'
import { EvaluationService } from './domain/evaluation'
import { CSVMaker, DefaultCSVMaker } from './core/csv'
import { FSFileWriter } from './core/file'
import { DefaultDateProvider } from './core/date'

export class EvalAssist {
  readonly project: ProjectService
  readonly member: MemberService
  readonly git: GitService
  readonly feedback: FeedbackService
  readonly evaluation: EvaluationService
  readonly docker: DockerService
  readonly csvmaker: CSVMaker

  constructor(config: Partial<EvalAssistConfig> = {}) {
    // Client
    const client = new AxiosHTTPClient(config.client)

    // Core
    const storageProvider = new DefaultStorageProvider()
    this.csvmaker = new DefaultCSVMaker()

    // Services
    this.member = new MemberService(client, storageProvider)
    this.project = new ProjectService(client, storageProvider)
    this.feedback = new FeedbackService(client, storageProvider)
    this.git = new GitService(client, this.project)
    this.docker = new DockerService(storageProvider, this.project, new DefaultDateProvider())
    this.evaluation = new EvaluationService(
      storageProvider,
      this.project,
      this.member,
      this.csvmaker,
      new FSFileWriter(),
      new DefaultDateProvider(),
    )

    // Hooks
    this.setupProjectInfosWriter(storageProvider)
  }

  private setupProjectInfosWriter(storageProvider: StorageProvider) {
    const writer = new ProjectDataCollectWriter(storageProvider, [
      // Project Datas Providers
      new ProjectDataProvider(this.project),
      new ProjectMembersDataProvider(this.member),
    ])

    // Write after clone
    this.git.addListener({
      onProjectClone: (params) => writer.write(params),
      onProjectPull: (params) => writer.write(params),
    })
  }
}

interface EvalAssistConfig {
  /**
   * Client Configuration
   */
  client: Partial<HTTPClientConfig>
}
